// Copyright IBM Corp. 2014, 2026
// SPDX-License-Identifier: BUSL-1.1

// Package schemarepo deals with the cross-cutting concern of gathering together
// all of the schemas needed for manipulating data produced by the various
// kinds of plugins.
package schemarepo
