// Copyright IBM Corp. 2014, 2026
// SPDX-License-Identifier: BUSL-1.1

// Package schema is a legacy package that used to represent the SDK, which is now its own
// library external to Terraform Core https://github.com/hashicorp/terraform-plugin-sdk
// Some of it is still used by Terraform's remote state backends, but this entire
// package should be removed in the future.
package schema
