// Copyright IBM Corp. 2014, 2026
// SPDX-License-Identifier: BUSL-1.1

// Package repl provides the structs and functions necessary to run
// REPL for Terraform. The REPL allows experimentation of Terraform
// interpolations without having to run a Terraform configuration.
package repl
