// Copyright IBM Corp. 2014, 2026
// SPDX-License-Identifier: BUSL-1.1

// Package statefile deals with the file format used to serialize states for
// persistent storage and then deserialize them into memory again later.
package statefile
