// Copyright IBM Corp. 2014, 2026
// SPDX-License-Identifier: BUSL-1.1

package main

import (
	"github.com/hashicorp/terraform/version"
)

var Version = version.Version

var VersionPrerelease = version.Prerelease
