// Code generated by "stringer -type ActionTriggerEvent"; DO NOT EDIT.

package configs

import "strconv"

func _() {
	// An "invalid array index" compiler error signifies that the constant values have changed.
	// Re-run the stringer command to generate them again.
	var x [1]struct{}
	_ = x[Unknown-0]
	_ = x[BeforeCreate-1]
	_ = x[AfterCreate-2]
	_ = x[BeforeUpdate-3]
	_ = x[AfterUpdate-4]
	_ = x[BeforeDestroy-5]
	_ = x[AfterDestroy-6]
	_ = x[Invoke-7]
}

const _ActionTriggerEvent_name = "UnknownBeforeCreateAfterCreateBeforeUpdateAfterUpdateBeforeDestroyAfterDestroyInvoke"

var _ActionTriggerEvent_index = [...]uint8{0, 7, 19, 30, 42, 53, 66, 78, 84}

func (i ActionTriggerEvent) String() string {
	idx := int(i) - 0
	if i < 0 || idx >= len(_ActionTriggerEvent_index)-1 {
		return "ActionTriggerEvent(" + strconv.FormatInt(int64(i), 10) + ")"
	}
	return _ActionTriggerEvent_name[_ActionTriggerEvent_index[idx]:_ActionTriggerEvent_index[idx+1]]
}
