// Copyright IBM Corp. 2014, 2026
// SPDX-License-Identifier: BUSL-1.1

// Package providers contains the interface and primary types required to
// implement a Terraform resource provider.
package providers
