// Copyright IBM Corp. 2014, 2026
// SPDX-License-Identifier: BUSL-1.1

// Package stackruntime contains the runtime implementation of the Stacks
// language, allowing the various operations over a stack configuration and
// its associated state and plans.
package stackruntime
