// Code generated by "stringer -type=ActionInvocationStatus resource_instance.go"; DO NOT EDIT.

package hooks

import "strconv"

func _() {
	// An "invalid array index" compiler error signifies that the constant values have changed.
	// Re-run the stringer command to generate them again.
	var x [1]struct{}
	_ = x[ActionInvocationStatusInvalid-0]
	_ = x[ActionInvocationPending-112]
	_ = x[ActionInvocationRunning-114]
	_ = x[ActionInvocationCompleted-67]
	_ = x[ActionInvocationErrored-69]
}

const (
	_ActionInvocationStatus_name_0 = "ActionInvocationStatusInvalid"
	_ActionInvocationStatus_name_1 = "ActionInvocationCompleted"
	_ActionInvocationStatus_name_2 = "ActionInvocationErrored"
	_ActionInvocationStatus_name_3 = "ActionInvocationPending"
	_ActionInvocationStatus_name_4 = "ActionInvocationRunning"
)

func (i ActionInvocationStatus) String() string {
	switch {
	case i == 0:
		return _ActionInvocationStatus_name_0
	case i == 67:
		return _ActionInvocationStatus_name_1
	case i == 69:
		return _ActionInvocationStatus_name_2
	case i == 112:
		return _ActionInvocationStatus_name_3
	case i == 114:
		return _ActionInvocationStatus_name_4
	default:
		return "ActionInvocationStatus(" + strconv.FormatInt(int64(i), 10) + ")"
	}
}
