// Copyright IBM Corp. 2014, 2026
// SPDX-License-Identifier: BUSL-1.1

// Package backendbase contains helpers for implementing the parts of remote
// state backends that tend to be treated similarly across all implementations.
package backendbase
