// Code generated by "stringer -type UnrecognizedKeyHandling"; DO NOT EDIT.

package statekeys

import "strconv"

func _() {
	// An "invalid array index" compiler error signifies that the constant values have changed.
	// Re-run the stringer command to generate them again.
	var x [1]struct{}
	_ = x[FailIfUnrecognized-70]
	_ = x[PreserveIfUnrecognized-80]
	_ = x[DiscardIfUnrecognized-68]
}

const (
	_UnrecognizedKeyHandling_name_0 = "DiscardIfUnrecognized"
	_UnrecognizedKeyHandling_name_1 = "FailIfUnrecognized"
	_UnrecognizedKeyHandling_name_2 = "PreserveIfUnrecognized"
)

func (i UnrecognizedKeyHandling) String() string {
	switch {
	case i == 68:
		return _UnrecognizedKeyHandling_name_0
	case i == 70:
		return _UnrecognizedKeyHandling_name_1
	case i == 80:
		return _UnrecognizedKeyHandling_name_2
	default:
		return "UnrecognizedKeyHandling(" + strconv.FormatInt(int64(i), 10) + ")"
	}
}
