// Copyright IBM Corp. 2014, 2026
// SPDX-License-Identifier: BUSL-1.1

// Package lang deals with the runtime aspects of Terraform's configuration
// language, with concerns such as expression evaluation. It is closely related
// to sibling package "configs", which is responsible for configuration
// parsing and static validation.
package lang
