// Copyright IBM Corp. 2014, 2026
// SPDX-License-Identifier: BUSL-1.1

// Package plans contains the types that are used to represent Terraform plans.
//
// A plan describes a set of changes that Terraform will make to update remote
// objects to match with changes to the configuration.
package plans
