// Copyright IBM Corp. 2014, 2026
// SPDX-License-Identifier: BUSL-1.1

// Package namedvals has container types to help with modelling the gradual
// evaluation of local values (input variable values, local values, output
// values) during a graph walk.
package namedvals
