// Copyright IBM Corp. 2014, 2026
// SPDX-License-Identifier: BUSL-1.1

// Package objchange deals with the business logic of taking a prior state
// value and a config value and producing a proposed new merged value, along
// with other related rules in this domain.
package objchange
