// Code generated by "stringer -type=Progress progress.go"; DO NOT EDIT.

package moduletest

import "strconv"

func _() {
	// An "invalid array index" compiler error signifies that the constant values have changed.
	// Re-run the stringer command to generate them again.
	var x [1]struct{}
	_ = x[Starting-0]
	_ = x[Running-1]
	_ = x[TearDown-2]
	_ = x[Complete-3]
}

const _Progress_name = "StartingRunningTearDownComplete"

var _Progress_index = [...]uint8{0, 8, 15, 23, 31}

func (i Progress) String() string {
	idx := int(i) - 0
	if i < 0 || idx >= len(_Progress_index)-1 {
		return "Progress(" + strconv.FormatInt(int64(i), 10) + ")"
	}
	return _Progress_name[_Progress_index[idx]:_Progress_index[idx+1]]
}
