// Copyright IBM Corp. 2014, 2026
// SPDX-License-Identifier: BUSL-1.1

// Code generated by protoc-gen-go. DO NOT EDIT.
// versions:
// 	protoc-gen-go v1.36.11
// 	protoc        v5.29.3
// source: terraform1.proto

package terraform1

import (
	protoreflect "google.golang.org/protobuf/reflect/protoreflect"
	protoimpl "google.golang.org/protobuf/runtime/protoimpl"
	reflect "reflect"
	sync "sync"
	unsafe "unsafe"
)

const (
	// Verify that this generated code is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(20 - protoimpl.MinVersion)
	// Verify that runtime/protoimpl is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(protoimpl.MaxVersion - 20)
)

type Diagnostic_Severity int32

const (
	Diagnostic_INVALID Diagnostic_Severity = 0
	Diagnostic_ERROR   Diagnostic_Severity = 1
	Diagnostic_WARNING Diagnostic_Severity = 2
)

// Enum value maps for Diagnostic_Severity.
var (
	Diagnostic_Severity_name = map[int32]string{
		0: "INVALID",
		1: "ERROR",
		2: "WARNING",
	}
	Diagnostic_Severity_value = map[string]int32{
		"INVALID": 0,
		"ERROR":   1,
		"WARNING": 2,
	}
)

func (x Diagnostic_Severity) Enum() *Diagnostic_Severity {
	p := new(Diagnostic_Severity)
	*p = x
	return p
}

func (x Diagnostic_Severity) String() string {
	return protoimpl.X.EnumStringOf(x.Descriptor(), protoreflect.EnumNumber(x))
}

func (Diagnostic_Severity) Descriptor() protoreflect.EnumDescriptor {
	return file_terraform1_proto_enumTypes[0].Descriptor()
}

func (Diagnostic_Severity) Type() protoreflect.EnumType {
	return &file_terraform1_proto_enumTypes[0]
}

func (x Diagnostic_Severity) Number() protoreflect.EnumNumber {
	return protoreflect.EnumNumber(x)
}

// Deprecated: Use Diagnostic_Severity.Descriptor instead.
func (Diagnostic_Severity) EnumDescriptor() ([]byte, []int) {
	return file_terraform1_proto_rawDescGZIP(), []int{2, 0}
}

// Represents a selected or available version of a provider, from either a
// dependency lock object (selected) or a provider cache object (available).
//
// This message type corresponds in meaning with a single "provider" block in a
// dependency lock file, but not all messages of this type directly represent
// such a physical block.
type ProviderPackage struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// The address of the provider using the canonical form of the provider
	// source address syntax.
	SourceAddr string `protobuf:"bytes,1,opt,name=source_addr,json=sourceAddr,proto3" json:"source_addr,omitempty"`
	// The version number of this provider package. Unset for (and only for)
	// built-in providers; callers may use the set-ness of this field to
	// distinguish installable vs. built-in providers without having to
	// parse the source address syntax.
	Version string `protobuf:"bytes,2,opt,name=version,proto3" json:"version,omitempty"`
	// The hash strings that Terraform knows about for this provider package,
	// using the same "scheme:hash" syntax used in Terraform's dependency
	// lock file format.
	//
	// For a message representing a "selected" provider package this enumerates
	// all of the checksums that were previously loaded from a dependency
	// lock file or otherwise inserted into a dependency locks object, which
	// are usually (but not necessarily) originally obtained from the
	// provider's origin registry and then cached in the lock file.
	//
	// For a message representing an "available" provider package this
	// describes only the actual package on disk, and so will typically
	// include only the subset of the checksums from the corresponding
	// "selected" package that are relevant to the current platform where
	// Terraform Core is running.
	Hashes        []string `protobuf:"bytes,3,rep,name=hashes,proto3" json:"hashes,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *ProviderPackage) Reset() {
	*x = ProviderPackage{}
	mi := &file_terraform1_proto_msgTypes[0]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *ProviderPackage) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*ProviderPackage) ProtoMessage() {}

func (x *ProviderPackage) ProtoReflect() protoreflect.Message {
	mi := &file_terraform1_proto_msgTypes[0]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use ProviderPackage.ProtoReflect.Descriptor instead.
func (*ProviderPackage) Descriptor() ([]byte, []int) {
	return file_terraform1_proto_rawDescGZIP(), []int{0}
}

func (x *ProviderPackage) GetSourceAddr() string {
	if x != nil {
		return x.SourceAddr
	}
	return ""
}

func (x *ProviderPackage) GetVersion() string {
	if x != nil {
		return x.Version
	}
	return ""
}

func (x *ProviderPackage) GetHashes() []string {
	if x != nil {
		return x.Hashes
	}
	return nil
}

// A source address in the same form as it would appear in a Terraform
// configuration: a source string combined with an optional version constraint
// string, where the latter is valid only for registry module addresses.
//
// This is not used for "final" source addresses that have already been reduced
// to an exact version selection. For those we just directly encode the string
// representation of the final address, including a version number if necessary.
type SourceAddress struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	Source        string                 `protobuf:"bytes,1,opt,name=source,proto3" json:"source,omitempty"`
	Versions      string                 `protobuf:"bytes,2,opt,name=versions,proto3" json:"versions,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *SourceAddress) Reset() {
	*x = SourceAddress{}
	mi := &file_terraform1_proto_msgTypes[1]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *SourceAddress) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*SourceAddress) ProtoMessage() {}

func (x *SourceAddress) ProtoReflect() protoreflect.Message {
	mi := &file_terraform1_proto_msgTypes[1]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use SourceAddress.ProtoReflect.Descriptor instead.
func (*SourceAddress) Descriptor() ([]byte, []int) {
	return file_terraform1_proto_rawDescGZIP(), []int{1}
}

func (x *SourceAddress) GetSource() string {
	if x != nil {
		return x.Source
	}
	return ""
}

func (x *SourceAddress) GetVersions() string {
	if x != nil {
		return x.Versions
	}
	return ""
}

type Diagnostic struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	Severity      Diagnostic_Severity    `protobuf:"varint,1,opt,name=severity,proto3,enum=terraform1.Diagnostic_Severity" json:"severity,omitempty"`
	Summary       string                 `protobuf:"bytes,2,opt,name=summary,proto3" json:"summary,omitempty"`
	Detail        string                 `protobuf:"bytes,3,opt,name=detail,proto3" json:"detail,omitempty"`
	Subject       *SourceRange           `protobuf:"bytes,4,opt,name=subject,proto3" json:"subject,omitempty"`
	Context       *SourceRange           `protobuf:"bytes,5,opt,name=context,proto3" json:"context,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *Diagnostic) Reset() {
	*x = Diagnostic{}
	mi := &file_terraform1_proto_msgTypes[2]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *Diagnostic) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*Diagnostic) ProtoMessage() {}

func (x *Diagnostic) ProtoReflect() protoreflect.Message {
	mi := &file_terraform1_proto_msgTypes[2]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use Diagnostic.ProtoReflect.Descriptor instead.
func (*Diagnostic) Descriptor() ([]byte, []int) {
	return file_terraform1_proto_rawDescGZIP(), []int{2}
}

func (x *Diagnostic) GetSeverity() Diagnostic_Severity {
	if x != nil {
		return x.Severity
	}
	return Diagnostic_INVALID
}

func (x *Diagnostic) GetSummary() string {
	if x != nil {
		return x.Summary
	}
	return ""
}

func (x *Diagnostic) GetDetail() string {
	if x != nil {
		return x.Detail
	}
	return ""
}

func (x *Diagnostic) GetSubject() *SourceRange {
	if x != nil {
		return x.Subject
	}
	return nil
}

func (x *Diagnostic) GetContext() *SourceRange {
	if x != nil {
		return x.Context
	}
	return nil
}

type SourceRange struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	SourceAddr    string                 `protobuf:"bytes,1,opt,name=source_addr,json=sourceAddr,proto3" json:"source_addr,omitempty"`
	Start         *SourcePos             `protobuf:"bytes,2,opt,name=start,proto3" json:"start,omitempty"`
	End           *SourcePos             `protobuf:"bytes,3,opt,name=end,proto3" json:"end,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *SourceRange) Reset() {
	*x = SourceRange{}
	mi := &file_terraform1_proto_msgTypes[3]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *SourceRange) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*SourceRange) ProtoMessage() {}

func (x *SourceRange) ProtoReflect() protoreflect.Message {
	mi := &file_terraform1_proto_msgTypes[3]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use SourceRange.ProtoReflect.Descriptor instead.
func (*SourceRange) Descriptor() ([]byte, []int) {
	return file_terraform1_proto_rawDescGZIP(), []int{3}
}

func (x *SourceRange) GetSourceAddr() string {
	if x != nil {
		return x.SourceAddr
	}
	return ""
}

func (x *SourceRange) GetStart() *SourcePos {
	if x != nil {
		return x.Start
	}
	return nil
}

func (x *SourceRange) GetEnd() *SourcePos {
	if x != nil {
		return x.End
	}
	return nil
}

type SourcePos struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	Byte          int64                  `protobuf:"varint,1,opt,name=byte,proto3" json:"byte,omitempty"`
	Line          int64                  `protobuf:"varint,2,opt,name=line,proto3" json:"line,omitempty"`
	Column        int64                  `protobuf:"varint,3,opt,name=column,proto3" json:"column,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *SourcePos) Reset() {
	*x = SourcePos{}
	mi := &file_terraform1_proto_msgTypes[4]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *SourcePos) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*SourcePos) ProtoMessage() {}

func (x *SourcePos) ProtoReflect() protoreflect.Message {
	mi := &file_terraform1_proto_msgTypes[4]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use SourcePos.ProtoReflect.Descriptor instead.
func (*SourcePos) Descriptor() ([]byte, []int) {
	return file_terraform1_proto_rawDescGZIP(), []int{4}
}

func (x *SourcePos) GetByte() int64 {
	if x != nil {
		return x.Byte
	}
	return 0
}

func (x *SourcePos) GetLine() int64 {
	if x != nil {
		return x.Line
	}
	return 0
}

func (x *SourcePos) GetColumn() int64 {
	if x != nil {
		return x.Column
	}
	return 0
}

var File_terraform1_proto protoreflect.FileDescriptor

const file_terraform1_proto_rawDesc = "" +
	"\n" +
	"\x10terraform1.proto\x12\n" +
	"terraform1\"d\n" +
	"\x0fProviderPackage\x12\x1f\n" +
	"\vsource_addr\x18\x01 \x01(\tR\n" +
	"sourceAddr\x12\x18\n" +
	"\aversion\x18\x02 \x01(\tR\aversion\x12\x16\n" +
	"\x06hashes\x18\x03 \x03(\tR\x06hashes\"C\n" +
	"\rSourceAddress\x12\x16\n" +
	"\x06source\x18\x01 \x01(\tR\x06source\x12\x1a\n" +
	"\bversions\x18\x02 \x01(\tR\bversions\"\x92\x02\n" +
	"\n" +
	"Diagnostic\x12;\n" +
	"\bseverity\x18\x01 \x01(\x0e2\x1f.terraform1.Diagnostic.SeverityR\bseverity\x12\x18\n" +
	"\asummary\x18\x02 \x01(\tR\asummary\x12\x16\n" +
	"\x06detail\x18\x03 \x01(\tR\x06detail\x121\n" +
	"\asubject\x18\x04 \x01(\v2\x17.terraform1.SourceRangeR\asubject\x121\n" +
	"\acontext\x18\x05 \x01(\v2\x17.terraform1.SourceRangeR\acontext\"/\n" +
	"\bSeverity\x12\v\n" +
	"\aINVALID\x10\x00\x12\t\n" +
	"\x05ERROR\x10\x01\x12\v\n" +
	"\aWARNING\x10\x02\"\x84\x01\n" +
	"\vSourceRange\x12\x1f\n" +
	"\vsource_addr\x18\x01 \x01(\tR\n" +
	"sourceAddr\x12+\n" +
	"\x05start\x18\x02 \x01(\v2\x15.terraform1.SourcePosR\x05start\x12'\n" +
	"\x03end\x18\x03 \x01(\v2\x15.terraform1.SourcePosR\x03end\"K\n" +
	"\tSourcePos\x12\x12\n" +
	"\x04byte\x18\x01 \x01(\x03R\x04byte\x12\x12\n" +
	"\x04line\x18\x02 \x01(\x03R\x04line\x12\x16\n" +
	"\x06column\x18\x03 \x01(\x03R\x06columnb\x06proto3"

var (
	file_terraform1_proto_rawDescOnce sync.Once
	file_terraform1_proto_rawDescData []byte
)

func file_terraform1_proto_rawDescGZIP() []byte {
	file_terraform1_proto_rawDescOnce.Do(func() {
		file_terraform1_proto_rawDescData = protoimpl.X.CompressGZIP(unsafe.Slice(unsafe.StringData(file_terraform1_proto_rawDesc), len(file_terraform1_proto_rawDesc)))
	})
	return file_terraform1_proto_rawDescData
}

var file_terraform1_proto_enumTypes = make([]protoimpl.EnumInfo, 1)
var file_terraform1_proto_msgTypes = make([]protoimpl.MessageInfo, 5)
var file_terraform1_proto_goTypes = []any{
	(Diagnostic_Severity)(0), // 0: terraform1.Diagnostic.Severity
	(*ProviderPackage)(nil),  // 1: terraform1.ProviderPackage
	(*SourceAddress)(nil),    // 2: terraform1.SourceAddress
	(*Diagnostic)(nil),       // 3: terraform1.Diagnostic
	(*SourceRange)(nil),      // 4: terraform1.SourceRange
	(*SourcePos)(nil),        // 5: terraform1.SourcePos
}
var file_terraform1_proto_depIdxs = []int32{
	0, // 0: terraform1.Diagnostic.severity:type_name -> terraform1.Diagnostic.Severity
	4, // 1: terraform1.Diagnostic.subject:type_name -> terraform1.SourceRange
	4, // 2: terraform1.Diagnostic.context:type_name -> terraform1.SourceRange
	5, // 3: terraform1.SourceRange.start:type_name -> terraform1.SourcePos
	5, // 4: terraform1.SourceRange.end:type_name -> terraform1.SourcePos
	5, // [5:5] is the sub-list for method output_type
	5, // [5:5] is the sub-list for method input_type
	5, // [5:5] is the sub-list for extension type_name
	5, // [5:5] is the sub-list for extension extendee
	0, // [0:5] is the sub-list for field type_name
}

func init() { file_terraform1_proto_init() }
func file_terraform1_proto_init() {
	if File_terraform1_proto != nil {
		return
	}
	type x struct{}
	out := protoimpl.TypeBuilder{
		File: protoimpl.DescBuilder{
			GoPackagePath: reflect.TypeOf(x{}).PkgPath(),
			RawDescriptor: unsafe.Slice(unsafe.StringData(file_terraform1_proto_rawDesc), len(file_terraform1_proto_rawDesc)),
			NumEnums:      1,
			NumMessages:   5,
			NumExtensions: 0,
			NumServices:   0,
		},
		GoTypes:           file_terraform1_proto_goTypes,
		DependencyIndexes: file_terraform1_proto_depIdxs,
		EnumInfos:         file_terraform1_proto_enumTypes,
		MessageInfos:      file_terraform1_proto_msgTypes,
	}.Build()
	File_terraform1_proto = out.File
	file_terraform1_proto_goTypes = nil
	file_terraform1_proto_depIdxs = nil
}
