// Copyright IBM Corp. 2014, 2026
// SPDX-License-Identifier: BUSL-1.1

// Package blocktoattr includes some helper functions that can perform
// preprocessing on a HCL body where a configschema.Block schema is available
// in order to allow list and set attributes defined in the schema to be
// optionally written by the user as block syntax.
package blocktoattr
