// Copyright IBM Corp. 2014, 2026
// SPDX-License-Identifier: BUSL-1.1

// Package modsdir is an internal package containing the model types used to
// represent the manifest of modules in a local modules cache directory.
package modsdir
