// Copyright IBM Corp. 2014, 2026
// SPDX-License-Identifier: BUSL-1.1

// Package stackaddrs builds on the top-level "addrs" package to provide the
// addresses for the extra layer of concepts that the stacks configuration
// language and its runtime are concerned about.
package stackaddrs
