// Copyright IBM Corp. 2014, 2026
// SPDX-License-Identifier: BUSL-1.1

// Package planproto is home to the Go stubs generated from the tfplan protobuf
// schema.
//
// This is an internal package to be used only by Terraform's planfile package.
// From elsewhere in Terraform, use the API exported by the planfile package
// itself.
package planproto
