// Code generated by "stringer -type Action"; DO NOT EDIT.

package plans

import "strconv"

func _() {
	// An "invalid array index" compiler error signifies that the constant values have changed.
	// Re-run the stringer command to generate them again.
	var x [1]struct{}
	_ = x[NoOp-0]
	_ = x[Create-43]
	_ = x[Read-8592]
	_ = x[Update-126]
	_ = x[DeleteThenCreate-8723]
	_ = x[CreateThenDelete-177]
	_ = x[Delete-45]
	_ = x[Forget-46]
	_ = x[CreateThenForget-10789]
	_ = x[Open-10179]
	_ = x[Renew-10227]
	_ = x[Close-10959]
}

const _Action_name = "NoOpCreateDeleteForgetUpdateCreateThenDeleteReadDeleteThenCreateOpenRenewCreateThenForgetClose"

var _Action_map = map[Action]string{
	0:     _Action_name[0:4],
	43:    _Action_name[4:10],
	45:    _Action_name[10:16],
	46:    _Action_name[16:22],
	126:   _Action_name[22:28],
	177:   _Action_name[28:44],
	8592:  _Action_name[44:48],
	8723:  _Action_name[48:64],
	10179: _Action_name[64:68],
	10227: _Action_name[68:73],
	10789: _Action_name[73:89],
	10959: _Action_name[89:94],
}

func (i Action) String() string {
	if str, ok := _Action_map[i]; ok {
		return str
	}
	return "Action(" + strconv.FormatInt(int64(i), 10) + ")"
}
