// Copyright IBM Corp. 2014, 2026
// SPDX-License-Identifier: BUSL-1.1

// Code generated by protoc-gen-go. DO NOT EDIT.
// versions:
// 	protoc-gen-go v1.36.11
// 	protoc        v5.29.3
// source: packages.proto

package packages

import (
	terraform1 "github.com/hashicorp/terraform/internal/rpcapi/terraform1"
	protoreflect "google.golang.org/protobuf/reflect/protoreflect"
	protoimpl "google.golang.org/protobuf/runtime/protoimpl"
	reflect "reflect"
	sync "sync"
	unsafe "unsafe"
)

const (
	// Verify that this generated code is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(20 - protoimpl.MinVersion)
	// Verify that runtime/protoimpl is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(protoimpl.MaxVersion - 20)
)

type ProviderPackageVersions struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *ProviderPackageVersions) Reset() {
	*x = ProviderPackageVersions{}
	mi := &file_packages_proto_msgTypes[0]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *ProviderPackageVersions) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*ProviderPackageVersions) ProtoMessage() {}

func (x *ProviderPackageVersions) ProtoReflect() protoreflect.Message {
	mi := &file_packages_proto_msgTypes[0]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use ProviderPackageVersions.ProtoReflect.Descriptor instead.
func (*ProviderPackageVersions) Descriptor() ([]byte, []int) {
	return file_packages_proto_rawDescGZIP(), []int{0}
}

type FetchProviderPackage struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *FetchProviderPackage) Reset() {
	*x = FetchProviderPackage{}
	mi := &file_packages_proto_msgTypes[1]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *FetchProviderPackage) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*FetchProviderPackage) ProtoMessage() {}

func (x *FetchProviderPackage) ProtoReflect() protoreflect.Message {
	mi := &file_packages_proto_msgTypes[1]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use FetchProviderPackage.ProtoReflect.Descriptor instead.
func (*FetchProviderPackage) Descriptor() ([]byte, []int) {
	return file_packages_proto_rawDescGZIP(), []int{1}
}

type ModulePackageVersions struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *ModulePackageVersions) Reset() {
	*x = ModulePackageVersions{}
	mi := &file_packages_proto_msgTypes[2]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *ModulePackageVersions) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*ModulePackageVersions) ProtoMessage() {}

func (x *ModulePackageVersions) ProtoReflect() protoreflect.Message {
	mi := &file_packages_proto_msgTypes[2]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use ModulePackageVersions.ProtoReflect.Descriptor instead.
func (*ModulePackageVersions) Descriptor() ([]byte, []int) {
	return file_packages_proto_rawDescGZIP(), []int{2}
}

type ModulePackageSourceAddr struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *ModulePackageSourceAddr) Reset() {
	*x = ModulePackageSourceAddr{}
	mi := &file_packages_proto_msgTypes[3]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *ModulePackageSourceAddr) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*ModulePackageSourceAddr) ProtoMessage() {}

func (x *ModulePackageSourceAddr) ProtoReflect() protoreflect.Message {
	mi := &file_packages_proto_msgTypes[3]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use ModulePackageSourceAddr.ProtoReflect.Descriptor instead.
func (*ModulePackageSourceAddr) Descriptor() ([]byte, []int) {
	return file_packages_proto_rawDescGZIP(), []int{3}
}

type FetchModulePackage struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *FetchModulePackage) Reset() {
	*x = FetchModulePackage{}
	mi := &file_packages_proto_msgTypes[4]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *FetchModulePackage) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*FetchModulePackage) ProtoMessage() {}

func (x *FetchModulePackage) ProtoReflect() protoreflect.Message {
	mi := &file_packages_proto_msgTypes[4]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use FetchModulePackage.ProtoReflect.Descriptor instead.
func (*FetchModulePackage) Descriptor() ([]byte, []int) {
	return file_packages_proto_rawDescGZIP(), []int{4}
}

type ProviderPackageVersions_Request struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	SourceAddr    string                 `protobuf:"bytes,1,opt,name=source_addr,json=sourceAddr,proto3" json:"source_addr,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *ProviderPackageVersions_Request) Reset() {
	*x = ProviderPackageVersions_Request{}
	mi := &file_packages_proto_msgTypes[5]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *ProviderPackageVersions_Request) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*ProviderPackageVersions_Request) ProtoMessage() {}

func (x *ProviderPackageVersions_Request) ProtoReflect() protoreflect.Message {
	mi := &file_packages_proto_msgTypes[5]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use ProviderPackageVersions_Request.ProtoReflect.Descriptor instead.
func (*ProviderPackageVersions_Request) Descriptor() ([]byte, []int) {
	return file_packages_proto_rawDescGZIP(), []int{0, 0}
}

func (x *ProviderPackageVersions_Request) GetSourceAddr() string {
	if x != nil {
		return x.SourceAddr
	}
	return ""
}

type ProviderPackageVersions_Response struct {
	state         protoimpl.MessageState   `protogen:"open.v1"`
	Versions      []string                 `protobuf:"bytes,1,rep,name=versions,proto3" json:"versions,omitempty"`
	Diagnostics   []*terraform1.Diagnostic `protobuf:"bytes,2,rep,name=diagnostics,proto3" json:"diagnostics,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *ProviderPackageVersions_Response) Reset() {
	*x = ProviderPackageVersions_Response{}
	mi := &file_packages_proto_msgTypes[6]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *ProviderPackageVersions_Response) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*ProviderPackageVersions_Response) ProtoMessage() {}

func (x *ProviderPackageVersions_Response) ProtoReflect() protoreflect.Message {
	mi := &file_packages_proto_msgTypes[6]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use ProviderPackageVersions_Response.ProtoReflect.Descriptor instead.
func (*ProviderPackageVersions_Response) Descriptor() ([]byte, []int) {
	return file_packages_proto_rawDescGZIP(), []int{0, 1}
}

func (x *ProviderPackageVersions_Response) GetVersions() []string {
	if x != nil {
		return x.Versions
	}
	return nil
}

func (x *ProviderPackageVersions_Response) GetDiagnostics() []*terraform1.Diagnostic {
	if x != nil {
		return x.Diagnostics
	}
	return nil
}

type FetchProviderPackage_Request struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	CacheDir      string                 `protobuf:"bytes,1,opt,name=cache_dir,json=cacheDir,proto3" json:"cache_dir,omitempty"`
	SourceAddr    string                 `protobuf:"bytes,2,opt,name=source_addr,json=sourceAddr,proto3" json:"source_addr,omitempty"`
	Version       string                 `protobuf:"bytes,3,opt,name=version,proto3" json:"version,omitempty"`
	Platforms     []string               `protobuf:"bytes,4,rep,name=platforms,proto3" json:"platforms,omitempty"`
	Hashes        []string               `protobuf:"bytes,5,rep,name=hashes,proto3" json:"hashes,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *FetchProviderPackage_Request) Reset() {
	*x = FetchProviderPackage_Request{}
	mi := &file_packages_proto_msgTypes[7]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *FetchProviderPackage_Request) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*FetchProviderPackage_Request) ProtoMessage() {}

func (x *FetchProviderPackage_Request) ProtoReflect() protoreflect.Message {
	mi := &file_packages_proto_msgTypes[7]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use FetchProviderPackage_Request.ProtoReflect.Descriptor instead.
func (*FetchProviderPackage_Request) Descriptor() ([]byte, []int) {
	return file_packages_proto_rawDescGZIP(), []int{1, 0}
}

func (x *FetchProviderPackage_Request) GetCacheDir() string {
	if x != nil {
		return x.CacheDir
	}
	return ""
}

func (x *FetchProviderPackage_Request) GetSourceAddr() string {
	if x != nil {
		return x.SourceAddr
	}
	return ""
}

func (x *FetchProviderPackage_Request) GetVersion() string {
	if x != nil {
		return x.Version
	}
	return ""
}

func (x *FetchProviderPackage_Request) GetPlatforms() []string {
	if x != nil {
		return x.Platforms
	}
	return nil
}

func (x *FetchProviderPackage_Request) GetHashes() []string {
	if x != nil {
		return x.Hashes
	}
	return nil
}

type FetchProviderPackage_Response struct {
	state protoimpl.MessageState `protogen:"open.v1"`
	// Each requested platform will return a result in this list. The order
	// of the returned results will match the order of the requested
	// platforms. If the binary for a given platform could not be downloaded
	// there will still be an entry in the results with diagnostics
	// explaining why.
	Results       []*FetchProviderPackage_PlatformResult `protobuf:"bytes,1,rep,name=results,proto3" json:"results,omitempty"`
	Diagnostics   []*terraform1.Diagnostic               `protobuf:"bytes,2,rep,name=diagnostics,proto3" json:"diagnostics,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *FetchProviderPackage_Response) Reset() {
	*x = FetchProviderPackage_Response{}
	mi := &file_packages_proto_msgTypes[8]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *FetchProviderPackage_Response) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*FetchProviderPackage_Response) ProtoMessage() {}

func (x *FetchProviderPackage_Response) ProtoReflect() protoreflect.Message {
	mi := &file_packages_proto_msgTypes[8]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use FetchProviderPackage_Response.ProtoReflect.Descriptor instead.
func (*FetchProviderPackage_Response) Descriptor() ([]byte, []int) {
	return file_packages_proto_rawDescGZIP(), []int{1, 1}
}

func (x *FetchProviderPackage_Response) GetResults() []*FetchProviderPackage_PlatformResult {
	if x != nil {
		return x.Results
	}
	return nil
}

func (x *FetchProviderPackage_Response) GetDiagnostics() []*terraform1.Diagnostic {
	if x != nil {
		return x.Diagnostics
	}
	return nil
}

type FetchProviderPackage_PlatformResult struct {
	state         protoimpl.MessageState      `protogen:"open.v1"`
	Provider      *terraform1.ProviderPackage `protobuf:"bytes,1,opt,name=provider,proto3" json:"provider,omitempty"`
	Diagnostics   []*terraform1.Diagnostic    `protobuf:"bytes,2,rep,name=diagnostics,proto3" json:"diagnostics,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *FetchProviderPackage_PlatformResult) Reset() {
	*x = FetchProviderPackage_PlatformResult{}
	mi := &file_packages_proto_msgTypes[9]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *FetchProviderPackage_PlatformResult) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*FetchProviderPackage_PlatformResult) ProtoMessage() {}

func (x *FetchProviderPackage_PlatformResult) ProtoReflect() protoreflect.Message {
	mi := &file_packages_proto_msgTypes[9]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use FetchProviderPackage_PlatformResult.ProtoReflect.Descriptor instead.
func (*FetchProviderPackage_PlatformResult) Descriptor() ([]byte, []int) {
	return file_packages_proto_rawDescGZIP(), []int{1, 2}
}

func (x *FetchProviderPackage_PlatformResult) GetProvider() *terraform1.ProviderPackage {
	if x != nil {
		return x.Provider
	}
	return nil
}

func (x *FetchProviderPackage_PlatformResult) GetDiagnostics() []*terraform1.Diagnostic {
	if x != nil {
		return x.Diagnostics
	}
	return nil
}

type ModulePackageVersions_Request struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	SourceAddr    string                 `protobuf:"bytes,2,opt,name=source_addr,json=sourceAddr,proto3" json:"source_addr,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *ModulePackageVersions_Request) Reset() {
	*x = ModulePackageVersions_Request{}
	mi := &file_packages_proto_msgTypes[10]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *ModulePackageVersions_Request) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*ModulePackageVersions_Request) ProtoMessage() {}

func (x *ModulePackageVersions_Request) ProtoReflect() protoreflect.Message {
	mi := &file_packages_proto_msgTypes[10]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use ModulePackageVersions_Request.ProtoReflect.Descriptor instead.
func (*ModulePackageVersions_Request) Descriptor() ([]byte, []int) {
	return file_packages_proto_rawDescGZIP(), []int{2, 0}
}

func (x *ModulePackageVersions_Request) GetSourceAddr() string {
	if x != nil {
		return x.SourceAddr
	}
	return ""
}

type ModulePackageVersions_Response struct {
	state         protoimpl.MessageState   `protogen:"open.v1"`
	Versions      []string                 `protobuf:"bytes,1,rep,name=versions,proto3" json:"versions,omitempty"`
	Diagnostics   []*terraform1.Diagnostic `protobuf:"bytes,2,rep,name=diagnostics,proto3" json:"diagnostics,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *ModulePackageVersions_Response) Reset() {
	*x = ModulePackageVersions_Response{}
	mi := &file_packages_proto_msgTypes[11]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *ModulePackageVersions_Response) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*ModulePackageVersions_Response) ProtoMessage() {}

func (x *ModulePackageVersions_Response) ProtoReflect() protoreflect.Message {
	mi := &file_packages_proto_msgTypes[11]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use ModulePackageVersions_Response.ProtoReflect.Descriptor instead.
func (*ModulePackageVersions_Response) Descriptor() ([]byte, []int) {
	return file_packages_proto_rawDescGZIP(), []int{2, 1}
}

func (x *ModulePackageVersions_Response) GetVersions() []string {
	if x != nil {
		return x.Versions
	}
	return nil
}

func (x *ModulePackageVersions_Response) GetDiagnostics() []*terraform1.Diagnostic {
	if x != nil {
		return x.Diagnostics
	}
	return nil
}

type ModulePackageSourceAddr_Request struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	SourceAddr    string                 `protobuf:"bytes,1,opt,name=source_addr,json=sourceAddr,proto3" json:"source_addr,omitempty"`
	Version       string                 `protobuf:"bytes,2,opt,name=version,proto3" json:"version,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *ModulePackageSourceAddr_Request) Reset() {
	*x = ModulePackageSourceAddr_Request{}
	mi := &file_packages_proto_msgTypes[12]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *ModulePackageSourceAddr_Request) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*ModulePackageSourceAddr_Request) ProtoMessage() {}

func (x *ModulePackageSourceAddr_Request) ProtoReflect() protoreflect.Message {
	mi := &file_packages_proto_msgTypes[12]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use ModulePackageSourceAddr_Request.ProtoReflect.Descriptor instead.
func (*ModulePackageSourceAddr_Request) Descriptor() ([]byte, []int) {
	return file_packages_proto_rawDescGZIP(), []int{3, 0}
}

func (x *ModulePackageSourceAddr_Request) GetSourceAddr() string {
	if x != nil {
		return x.SourceAddr
	}
	return ""
}

func (x *ModulePackageSourceAddr_Request) GetVersion() string {
	if x != nil {
		return x.Version
	}
	return ""
}

type ModulePackageSourceAddr_Response struct {
	state         protoimpl.MessageState   `protogen:"open.v1"`
	Url           string                   `protobuf:"bytes,1,opt,name=url,proto3" json:"url,omitempty"`
	Diagnostics   []*terraform1.Diagnostic `protobuf:"bytes,2,rep,name=diagnostics,proto3" json:"diagnostics,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *ModulePackageSourceAddr_Response) Reset() {
	*x = ModulePackageSourceAddr_Response{}
	mi := &file_packages_proto_msgTypes[13]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *ModulePackageSourceAddr_Response) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*ModulePackageSourceAddr_Response) ProtoMessage() {}

func (x *ModulePackageSourceAddr_Response) ProtoReflect() protoreflect.Message {
	mi := &file_packages_proto_msgTypes[13]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use ModulePackageSourceAddr_Response.ProtoReflect.Descriptor instead.
func (*ModulePackageSourceAddr_Response) Descriptor() ([]byte, []int) {
	return file_packages_proto_rawDescGZIP(), []int{3, 1}
}

func (x *ModulePackageSourceAddr_Response) GetUrl() string {
	if x != nil {
		return x.Url
	}
	return ""
}

func (x *ModulePackageSourceAddr_Response) GetDiagnostics() []*terraform1.Diagnostic {
	if x != nil {
		return x.Diagnostics
	}
	return nil
}

type FetchModulePackage_Request struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	CacheDir      string                 `protobuf:"bytes,1,opt,name=cache_dir,json=cacheDir,proto3" json:"cache_dir,omitempty"`
	Url           string                 `protobuf:"bytes,2,opt,name=url,proto3" json:"url,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *FetchModulePackage_Request) Reset() {
	*x = FetchModulePackage_Request{}
	mi := &file_packages_proto_msgTypes[14]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *FetchModulePackage_Request) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*FetchModulePackage_Request) ProtoMessage() {}

func (x *FetchModulePackage_Request) ProtoReflect() protoreflect.Message {
	mi := &file_packages_proto_msgTypes[14]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use FetchModulePackage_Request.ProtoReflect.Descriptor instead.
func (*FetchModulePackage_Request) Descriptor() ([]byte, []int) {
	return file_packages_proto_rawDescGZIP(), []int{4, 0}
}

func (x *FetchModulePackage_Request) GetCacheDir() string {
	if x != nil {
		return x.CacheDir
	}
	return ""
}

func (x *FetchModulePackage_Request) GetUrl() string {
	if x != nil {
		return x.Url
	}
	return ""
}

type FetchModulePackage_Response struct {
	state         protoimpl.MessageState   `protogen:"open.v1"`
	Diagnostics   []*terraform1.Diagnostic `protobuf:"bytes,1,rep,name=diagnostics,proto3" json:"diagnostics,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *FetchModulePackage_Response) Reset() {
	*x = FetchModulePackage_Response{}
	mi := &file_packages_proto_msgTypes[15]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *FetchModulePackage_Response) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*FetchModulePackage_Response) ProtoMessage() {}

func (x *FetchModulePackage_Response) ProtoReflect() protoreflect.Message {
	mi := &file_packages_proto_msgTypes[15]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use FetchModulePackage_Response.ProtoReflect.Descriptor instead.
func (*FetchModulePackage_Response) Descriptor() ([]byte, []int) {
	return file_packages_proto_rawDescGZIP(), []int{4, 1}
}

func (x *FetchModulePackage_Response) GetDiagnostics() []*terraform1.Diagnostic {
	if x != nil {
		return x.Diagnostics
	}
	return nil
}

var File_packages_proto protoreflect.FileDescriptor

const file_packages_proto_rawDesc = "" +
	"\n" +
	"\x0epackages.proto\x12\x13terraform1.packages\x1a\x10terraform1.proto\"\xa7\x01\n" +
	"\x17ProviderPackageVersions\x1a*\n" +
	"\aRequest\x12\x1f\n" +
	"\vsource_addr\x18\x01 \x01(\tR\n" +
	"sourceAddr\x1a`\n" +
	"\bResponse\x12\x1a\n" +
	"\bversions\x18\x01 \x03(\tR\bversions\x128\n" +
	"\vdiagnostics\x18\x02 \x03(\v2\x16.terraform1.DiagnosticR\vdiagnostics\"\xd1\x03\n" +
	"\x14FetchProviderPackage\x1a\x97\x01\n" +
	"\aRequest\x12\x1b\n" +
	"\tcache_dir\x18\x01 \x01(\tR\bcacheDir\x12\x1f\n" +
	"\vsource_addr\x18\x02 \x01(\tR\n" +
	"sourceAddr\x12\x18\n" +
	"\aversion\x18\x03 \x01(\tR\aversion\x12\x1c\n" +
	"\tplatforms\x18\x04 \x03(\tR\tplatforms\x12\x16\n" +
	"\x06hashes\x18\x05 \x03(\tR\x06hashes\x1a\x98\x01\n" +
	"\bResponse\x12R\n" +
	"\aresults\x18\x01 \x03(\v28.terraform1.packages.FetchProviderPackage.PlatformResultR\aresults\x128\n" +
	"\vdiagnostics\x18\x02 \x03(\v2\x16.terraform1.DiagnosticR\vdiagnostics\x1a\x83\x01\n" +
	"\x0ePlatformResult\x127\n" +
	"\bprovider\x18\x01 \x01(\v2\x1b.terraform1.ProviderPackageR\bprovider\x128\n" +
	"\vdiagnostics\x18\x02 \x03(\v2\x16.terraform1.DiagnosticR\vdiagnostics\"\xa5\x01\n" +
	"\x15ModulePackageVersions\x1a*\n" +
	"\aRequest\x12\x1f\n" +
	"\vsource_addr\x18\x02 \x01(\tR\n" +
	"sourceAddr\x1a`\n" +
	"\bResponse\x12\x1a\n" +
	"\bversions\x18\x01 \x03(\tR\bversions\x128\n" +
	"\vdiagnostics\x18\x02 \x03(\v2\x16.terraform1.DiagnosticR\vdiagnostics\"\xb7\x01\n" +
	"\x17ModulePackageSourceAddr\x1aD\n" +
	"\aRequest\x12\x1f\n" +
	"\vsource_addr\x18\x01 \x01(\tR\n" +
	"sourceAddr\x12\x18\n" +
	"\aversion\x18\x02 \x01(\tR\aversion\x1aV\n" +
	"\bResponse\x12\x10\n" +
	"\x03url\x18\x01 \x01(\tR\x03url\x128\n" +
	"\vdiagnostics\x18\x02 \x03(\v2\x16.terraform1.DiagnosticR\vdiagnostics\"\x94\x01\n" +
	"\x12FetchModulePackage\x1a8\n" +
	"\aRequest\x12\x1b\n" +
	"\tcache_dir\x18\x01 \x01(\tR\bcacheDir\x12\x10\n" +
	"\x03url\x18\x02 \x01(\tR\x03url\x1aD\n" +
	"\bResponse\x128\n" +
	"\vdiagnostics\x18\x01 \x03(\v2\x16.terraform1.DiagnosticR\vdiagnostics2\x97\x05\n" +
	"\bPackages\x12\x86\x01\n" +
	"\x17ProviderPackageVersions\x124.terraform1.packages.ProviderPackageVersions.Request\x1a5.terraform1.packages.ProviderPackageVersions.Response\x12}\n" +
	"\x14FetchProviderPackage\x121.terraform1.packages.FetchProviderPackage.Request\x1a2.terraform1.packages.FetchProviderPackage.Response\x12\x80\x01\n" +
	"\x15ModulePackageVersions\x122.terraform1.packages.ModulePackageVersions.Request\x1a3.terraform1.packages.ModulePackageVersions.Response\x12\x86\x01\n" +
	"\x17ModulePackageSourceAddr\x124.terraform1.packages.ModulePackageSourceAddr.Request\x1a5.terraform1.packages.ModulePackageSourceAddr.Response\x12w\n" +
	"\x12FetchModulePackage\x12/.terraform1.packages.FetchModulePackage.Request\x1a0.terraform1.packages.FetchModulePackage.Responseb\x06proto3"

var (
	file_packages_proto_rawDescOnce sync.Once
	file_packages_proto_rawDescData []byte
)

func file_packages_proto_rawDescGZIP() []byte {
	file_packages_proto_rawDescOnce.Do(func() {
		file_packages_proto_rawDescData = protoimpl.X.CompressGZIP(unsafe.Slice(unsafe.StringData(file_packages_proto_rawDesc), len(file_packages_proto_rawDesc)))
	})
	return file_packages_proto_rawDescData
}

var file_packages_proto_msgTypes = make([]protoimpl.MessageInfo, 16)
var file_packages_proto_goTypes = []any{
	(*ProviderPackageVersions)(nil),             // 0: terraform1.packages.ProviderPackageVersions
	(*FetchProviderPackage)(nil),                // 1: terraform1.packages.FetchProviderPackage
	(*ModulePackageVersions)(nil),               // 2: terraform1.packages.ModulePackageVersions
	(*ModulePackageSourceAddr)(nil),             // 3: terraform1.packages.ModulePackageSourceAddr
	(*FetchModulePackage)(nil),                  // 4: terraform1.packages.FetchModulePackage
	(*ProviderPackageVersions_Request)(nil),     // 5: terraform1.packages.ProviderPackageVersions.Request
	(*ProviderPackageVersions_Response)(nil),    // 6: terraform1.packages.ProviderPackageVersions.Response
	(*FetchProviderPackage_Request)(nil),        // 7: terraform1.packages.FetchProviderPackage.Request
	(*FetchProviderPackage_Response)(nil),       // 8: terraform1.packages.FetchProviderPackage.Response
	(*FetchProviderPackage_PlatformResult)(nil), // 9: terraform1.packages.FetchProviderPackage.PlatformResult
	(*ModulePackageVersions_Request)(nil),       // 10: terraform1.packages.ModulePackageVersions.Request
	(*ModulePackageVersions_Response)(nil),      // 11: terraform1.packages.ModulePackageVersions.Response
	(*ModulePackageSourceAddr_Request)(nil),     // 12: terraform1.packages.ModulePackageSourceAddr.Request
	(*ModulePackageSourceAddr_Response)(nil),    // 13: terraform1.packages.ModulePackageSourceAddr.Response
	(*FetchModulePackage_Request)(nil),          // 14: terraform1.packages.FetchModulePackage.Request
	(*FetchModulePackage_Response)(nil),         // 15: terraform1.packages.FetchModulePackage.Response
	(*terraform1.Diagnostic)(nil),               // 16: terraform1.Diagnostic
	(*terraform1.ProviderPackage)(nil),          // 17: terraform1.ProviderPackage
}
var file_packages_proto_depIdxs = []int32{
	16, // 0: terraform1.packages.ProviderPackageVersions.Response.diagnostics:type_name -> terraform1.Diagnostic
	9,  // 1: terraform1.packages.FetchProviderPackage.Response.results:type_name -> terraform1.packages.FetchProviderPackage.PlatformResult
	16, // 2: terraform1.packages.FetchProviderPackage.Response.diagnostics:type_name -> terraform1.Diagnostic
	17, // 3: terraform1.packages.FetchProviderPackage.PlatformResult.provider:type_name -> terraform1.ProviderPackage
	16, // 4: terraform1.packages.FetchProviderPackage.PlatformResult.diagnostics:type_name -> terraform1.Diagnostic
	16, // 5: terraform1.packages.ModulePackageVersions.Response.diagnostics:type_name -> terraform1.Diagnostic
	16, // 6: terraform1.packages.ModulePackageSourceAddr.Response.diagnostics:type_name -> terraform1.Diagnostic
	16, // 7: terraform1.packages.FetchModulePackage.Response.diagnostics:type_name -> terraform1.Diagnostic
	5,  // 8: terraform1.packages.Packages.ProviderPackageVersions:input_type -> terraform1.packages.ProviderPackageVersions.Request
	7,  // 9: terraform1.packages.Packages.FetchProviderPackage:input_type -> terraform1.packages.FetchProviderPackage.Request
	10, // 10: terraform1.packages.Packages.ModulePackageVersions:input_type -> terraform1.packages.ModulePackageVersions.Request
	12, // 11: terraform1.packages.Packages.ModulePackageSourceAddr:input_type -> terraform1.packages.ModulePackageSourceAddr.Request
	14, // 12: terraform1.packages.Packages.FetchModulePackage:input_type -> terraform1.packages.FetchModulePackage.Request
	6,  // 13: terraform1.packages.Packages.ProviderPackageVersions:output_type -> terraform1.packages.ProviderPackageVersions.Response
	8,  // 14: terraform1.packages.Packages.FetchProviderPackage:output_type -> terraform1.packages.FetchProviderPackage.Response
	11, // 15: terraform1.packages.Packages.ModulePackageVersions:output_type -> terraform1.packages.ModulePackageVersions.Response
	13, // 16: terraform1.packages.Packages.ModulePackageSourceAddr:output_type -> terraform1.packages.ModulePackageSourceAddr.Response
	15, // 17: terraform1.packages.Packages.FetchModulePackage:output_type -> terraform1.packages.FetchModulePackage.Response
	13, // [13:18] is the sub-list for method output_type
	8,  // [8:13] is the sub-list for method input_type
	8,  // [8:8] is the sub-list for extension type_name
	8,  // [8:8] is the sub-list for extension extendee
	0,  // [0:8] is the sub-list for field type_name
}

func init() { file_packages_proto_init() }
func file_packages_proto_init() {
	if File_packages_proto != nil {
		return
	}
	type x struct{}
	out := protoimpl.TypeBuilder{
		File: protoimpl.DescBuilder{
			GoPackagePath: reflect.TypeOf(x{}).PkgPath(),
			RawDescriptor: unsafe.Slice(unsafe.StringData(file_packages_proto_rawDesc), len(file_packages_proto_rawDesc)),
			NumEnums:      0,
			NumMessages:   16,
			NumExtensions: 0,
			NumServices:   1,
		},
		GoTypes:           file_packages_proto_goTypes,
		DependencyIndexes: file_packages_proto_depIdxs,
		MessageInfos:      file_packages_proto_msgTypes,
	}.Build()
	File_packages_proto = out.File
	file_packages_proto_goTypes = nil
	file_packages_proto_depIdxs = nil
}
