// Copyright IBM Corp. 2014, 2026
// SPDX-License-Identifier: BUSL-1.1

// Code generated by protoc-gen-go. DO NOT EDIT.
// versions:
// 	protoc-gen-go v1.36.11
// 	protoc        v5.29.3
// source: setup.proto

package setup

import (
	protoreflect "google.golang.org/protobuf/reflect/protoreflect"
	protoimpl "google.golang.org/protobuf/runtime/protoimpl"
	reflect "reflect"
	sync "sync"
	unsafe "unsafe"
)

const (
	// Verify that this generated code is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(20 - protoimpl.MinVersion)
	// Verify that runtime/protoimpl is sufficiently up-to-date.
	_ = protoimpl.EnforceVersion(protoimpl.MaxVersion - 20)
)

type Handshake struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *Handshake) Reset() {
	*x = Handshake{}
	mi := &file_setup_proto_msgTypes[0]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *Handshake) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*Handshake) ProtoMessage() {}

func (x *Handshake) ProtoReflect() protoreflect.Message {
	mi := &file_setup_proto_msgTypes[0]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use Handshake.ProtoReflect.Descriptor instead.
func (*Handshake) Descriptor() ([]byte, []int) {
	return file_setup_proto_rawDescGZIP(), []int{0}
}

type Stop struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *Stop) Reset() {
	*x = Stop{}
	mi := &file_setup_proto_msgTypes[1]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *Stop) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*Stop) ProtoMessage() {}

func (x *Stop) ProtoReflect() protoreflect.Message {
	mi := &file_setup_proto_msgTypes[1]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use Stop.ProtoReflect.Descriptor instead.
func (*Stop) Descriptor() ([]byte, []int) {
	return file_setup_proto_rawDescGZIP(), []int{1}
}

type Config struct {
	state         protoimpl.MessageState     `protogen:"open.v1"`
	Credentials   map[string]*HostCredential `protobuf:"bytes,1,rep,name=credentials,proto3" json:"credentials,omitempty" protobuf_key:"bytes,1,opt,name=key" protobuf_val:"bytes,2,opt,name=value"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *Config) Reset() {
	*x = Config{}
	mi := &file_setup_proto_msgTypes[2]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *Config) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*Config) ProtoMessage() {}

func (x *Config) ProtoReflect() protoreflect.Message {
	mi := &file_setup_proto_msgTypes[2]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use Config.ProtoReflect.Descriptor instead.
func (*Config) Descriptor() ([]byte, []int) {
	return file_setup_proto_rawDescGZIP(), []int{2}
}

func (x *Config) GetCredentials() map[string]*HostCredential {
	if x != nil {
		return x.Credentials
	}
	return nil
}

type HostCredential struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	Token         string                 `protobuf:"bytes,1,opt,name=token,proto3" json:"token,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *HostCredential) Reset() {
	*x = HostCredential{}
	mi := &file_setup_proto_msgTypes[3]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *HostCredential) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*HostCredential) ProtoMessage() {}

func (x *HostCredential) ProtoReflect() protoreflect.Message {
	mi := &file_setup_proto_msgTypes[3]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use HostCredential.ProtoReflect.Descriptor instead.
func (*HostCredential) Descriptor() ([]byte, []int) {
	return file_setup_proto_rawDescGZIP(), []int{3}
}

func (x *HostCredential) GetToken() string {
	if x != nil {
		return x.Token
	}
	return ""
}

// The capabilities that the client wishes to advertise to the server during
// handshake.
type ClientCapabilities struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *ClientCapabilities) Reset() {
	*x = ClientCapabilities{}
	mi := &file_setup_proto_msgTypes[4]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *ClientCapabilities) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*ClientCapabilities) ProtoMessage() {}

func (x *ClientCapabilities) ProtoReflect() protoreflect.Message {
	mi := &file_setup_proto_msgTypes[4]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use ClientCapabilities.ProtoReflect.Descriptor instead.
func (*ClientCapabilities) Descriptor() ([]byte, []int) {
	return file_setup_proto_rawDescGZIP(), []int{4}
}

// The capabilities that the server wishes to advertise to the client during
// handshake. Fields in this message can also be used to acknowledge and
// confirm support for client capabilities advertised in ClientCapabilities,
// in situations where the client must vary its behavior based on the server's
// level of support.
type ServerCapabilities struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *ServerCapabilities) Reset() {
	*x = ServerCapabilities{}
	mi := &file_setup_proto_msgTypes[5]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *ServerCapabilities) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*ServerCapabilities) ProtoMessage() {}

func (x *ServerCapabilities) ProtoReflect() protoreflect.Message {
	mi := &file_setup_proto_msgTypes[5]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use ServerCapabilities.ProtoReflect.Descriptor instead.
func (*ServerCapabilities) Descriptor() ([]byte, []int) {
	return file_setup_proto_rawDescGZIP(), []int{5}
}

type Handshake_Request struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	Capabilities  *ClientCapabilities    `protobuf:"bytes,1,opt,name=capabilities,proto3" json:"capabilities,omitempty"`
	Config        *Config                `protobuf:"bytes,2,opt,name=config,proto3" json:"config,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *Handshake_Request) Reset() {
	*x = Handshake_Request{}
	mi := &file_setup_proto_msgTypes[6]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *Handshake_Request) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*Handshake_Request) ProtoMessage() {}

func (x *Handshake_Request) ProtoReflect() protoreflect.Message {
	mi := &file_setup_proto_msgTypes[6]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use Handshake_Request.ProtoReflect.Descriptor instead.
func (*Handshake_Request) Descriptor() ([]byte, []int) {
	return file_setup_proto_rawDescGZIP(), []int{0, 0}
}

func (x *Handshake_Request) GetCapabilities() *ClientCapabilities {
	if x != nil {
		return x.Capabilities
	}
	return nil
}

func (x *Handshake_Request) GetConfig() *Config {
	if x != nil {
		return x.Config
	}
	return nil
}

type Handshake_Response struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	Capabilities  *ServerCapabilities    `protobuf:"bytes,2,opt,name=capabilities,proto3" json:"capabilities,omitempty"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *Handshake_Response) Reset() {
	*x = Handshake_Response{}
	mi := &file_setup_proto_msgTypes[7]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *Handshake_Response) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*Handshake_Response) ProtoMessage() {}

func (x *Handshake_Response) ProtoReflect() protoreflect.Message {
	mi := &file_setup_proto_msgTypes[7]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use Handshake_Response.ProtoReflect.Descriptor instead.
func (*Handshake_Response) Descriptor() ([]byte, []int) {
	return file_setup_proto_rawDescGZIP(), []int{0, 1}
}

func (x *Handshake_Response) GetCapabilities() *ServerCapabilities {
	if x != nil {
		return x.Capabilities
	}
	return nil
}

type Stop_Request struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *Stop_Request) Reset() {
	*x = Stop_Request{}
	mi := &file_setup_proto_msgTypes[8]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *Stop_Request) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*Stop_Request) ProtoMessage() {}

func (x *Stop_Request) ProtoReflect() protoreflect.Message {
	mi := &file_setup_proto_msgTypes[8]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use Stop_Request.ProtoReflect.Descriptor instead.
func (*Stop_Request) Descriptor() ([]byte, []int) {
	return file_setup_proto_rawDescGZIP(), []int{1, 0}
}

type Stop_Response struct {
	state         protoimpl.MessageState `protogen:"open.v1"`
	unknownFields protoimpl.UnknownFields
	sizeCache     protoimpl.SizeCache
}

func (x *Stop_Response) Reset() {
	*x = Stop_Response{}
	mi := &file_setup_proto_msgTypes[9]
	ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
	ms.StoreMessageInfo(mi)
}

func (x *Stop_Response) String() string {
	return protoimpl.X.MessageStringOf(x)
}

func (*Stop_Response) ProtoMessage() {}

func (x *Stop_Response) ProtoReflect() protoreflect.Message {
	mi := &file_setup_proto_msgTypes[9]
	if x != nil {
		ms := protoimpl.X.MessageStateOf(protoimpl.Pointer(x))
		if ms.LoadMessageInfo() == nil {
			ms.StoreMessageInfo(mi)
		}
		return ms
	}
	return mi.MessageOf(x)
}

// Deprecated: Use Stop_Response.ProtoReflect.Descriptor instead.
func (*Stop_Response) Descriptor() ([]byte, []int) {
	return file_setup_proto_rawDescGZIP(), []int{1, 1}
}

var File_setup_proto protoreflect.FileDescriptor

const file_setup_proto_rawDesc = "" +
	"\n" +
	"\vsetup.proto\x12\x10terraform1.setup\"\xe9\x01\n" +
	"\tHandshake\x1a\x85\x01\n" +
	"\aRequest\x12H\n" +
	"\fcapabilities\x18\x01 \x01(\v2$.terraform1.setup.ClientCapabilitiesR\fcapabilities\x120\n" +
	"\x06config\x18\x02 \x01(\v2\x18.terraform1.setup.ConfigR\x06config\x1aT\n" +
	"\bResponse\x12H\n" +
	"\fcapabilities\x18\x02 \x01(\v2$.terraform1.setup.ServerCapabilitiesR\fcapabilities\"\x1d\n" +
	"\x04Stop\x1a\t\n" +
	"\aRequest\x1a\n" +
	"\n" +
	"\bResponse\"\xb7\x01\n" +
	"\x06Config\x12K\n" +
	"\vcredentials\x18\x01 \x03(\v2).terraform1.setup.Config.CredentialsEntryR\vcredentials\x1a`\n" +
	"\x10CredentialsEntry\x12\x10\n" +
	"\x03key\x18\x01 \x01(\tR\x03key\x126\n" +
	"\x05value\x18\x02 \x01(\v2 .terraform1.setup.HostCredentialR\x05value:\x028\x01\"&\n" +
	"\x0eHostCredential\x12\x14\n" +
	"\x05token\x18\x01 \x01(\tR\x05token\"\x14\n" +
	"\x12ClientCapabilities\"\x14\n" +
	"\x12ServerCapabilities2\xa8\x01\n" +
	"\x05Setup\x12V\n" +
	"\tHandshake\x12#.terraform1.setup.Handshake.Request\x1a$.terraform1.setup.Handshake.Response\x12G\n" +
	"\x04Stop\x12\x1e.terraform1.setup.Stop.Request\x1a\x1f.terraform1.setup.Stop.Responseb\x06proto3"

var (
	file_setup_proto_rawDescOnce sync.Once
	file_setup_proto_rawDescData []byte
)

func file_setup_proto_rawDescGZIP() []byte {
	file_setup_proto_rawDescOnce.Do(func() {
		file_setup_proto_rawDescData = protoimpl.X.CompressGZIP(unsafe.Slice(unsafe.StringData(file_setup_proto_rawDesc), len(file_setup_proto_rawDesc)))
	})
	return file_setup_proto_rawDescData
}

var file_setup_proto_msgTypes = make([]protoimpl.MessageInfo, 11)
var file_setup_proto_goTypes = []any{
	(*Handshake)(nil),          // 0: terraform1.setup.Handshake
	(*Stop)(nil),               // 1: terraform1.setup.Stop
	(*Config)(nil),             // 2: terraform1.setup.Config
	(*HostCredential)(nil),     // 3: terraform1.setup.HostCredential
	(*ClientCapabilities)(nil), // 4: terraform1.setup.ClientCapabilities
	(*ServerCapabilities)(nil), // 5: terraform1.setup.ServerCapabilities
	(*Handshake_Request)(nil),  // 6: terraform1.setup.Handshake.Request
	(*Handshake_Response)(nil), // 7: terraform1.setup.Handshake.Response
	(*Stop_Request)(nil),       // 8: terraform1.setup.Stop.Request
	(*Stop_Response)(nil),      // 9: terraform1.setup.Stop.Response
	nil,                        // 10: terraform1.setup.Config.CredentialsEntry
}
var file_setup_proto_depIdxs = []int32{
	10, // 0: terraform1.setup.Config.credentials:type_name -> terraform1.setup.Config.CredentialsEntry
	4,  // 1: terraform1.setup.Handshake.Request.capabilities:type_name -> terraform1.setup.ClientCapabilities
	2,  // 2: terraform1.setup.Handshake.Request.config:type_name -> terraform1.setup.Config
	5,  // 3: terraform1.setup.Handshake.Response.capabilities:type_name -> terraform1.setup.ServerCapabilities
	3,  // 4: terraform1.setup.Config.CredentialsEntry.value:type_name -> terraform1.setup.HostCredential
	6,  // 5: terraform1.setup.Setup.Handshake:input_type -> terraform1.setup.Handshake.Request
	8,  // 6: terraform1.setup.Setup.Stop:input_type -> terraform1.setup.Stop.Request
	7,  // 7: terraform1.setup.Setup.Handshake:output_type -> terraform1.setup.Handshake.Response
	9,  // 8: terraform1.setup.Setup.Stop:output_type -> terraform1.setup.Stop.Response
	7,  // [7:9] is the sub-list for method output_type
	5,  // [5:7] is the sub-list for method input_type
	5,  // [5:5] is the sub-list for extension type_name
	5,  // [5:5] is the sub-list for extension extendee
	0,  // [0:5] is the sub-list for field type_name
}

func init() { file_setup_proto_init() }
func file_setup_proto_init() {
	if File_setup_proto != nil {
		return
	}
	type x struct{}
	out := protoimpl.TypeBuilder{
		File: protoimpl.DescBuilder{
			GoPackagePath: reflect.TypeOf(x{}).PkgPath(),
			RawDescriptor: unsafe.Slice(unsafe.StringData(file_setup_proto_rawDesc), len(file_setup_proto_rawDesc)),
			NumEnums:      0,
			NumMessages:   11,
			NumExtensions: 0,
			NumServices:   1,
		},
		GoTypes:           file_setup_proto_goTypes,
		DependencyIndexes: file_setup_proto_depIdxs,
		MessageInfos:      file_setup_proto_msgTypes,
	}.Build()
	File_setup_proto = out.File
	file_setup_proto_goTypes = nil
	file_setup_proto_depIdxs = nil
}
