// Copyright IBM Corp. 2014, 2026
// SPDX-License-Identifier: BUSL-1.1

// Package jsonchecks implements the common JSON representation of check
// results/statuses that we use across both the JSON plan and JSON state
// representations.
package jsonchecks
