// Copyright IBM Corp. 2014, 2026
// SPDX-License-Identifier: BUSL-1.1

// Package provisioners contains the interface and primary types to implement a
// Terraform resource provisioner.
package provisioners
